#include <regdef.h>
#include <pal/palcalls.h>

	.text

	.globl	l4_alpha_ipc_call
	.globl	l4_alpha_ipc_send
	.globl	l4_alpha_ipc_wait
	.globl	l4_alpha_ipc_receive
	.globl	l4_alpha_ipc_reply_and_wait

	.set noreorder


#define pal	.long
		

#define save\
	subq	sp, 0x50, sp;\
	stq	gp, 0x00(sp);\
	stq	pv, 0x08(sp);\
	stq	ra, 0x10(sp);\
	stq	s0, 0x18(sp);\
	stq	s1, 0x20(sp);\
	stq	s2, 0x28(sp);\
	stq	s3, 0x30(sp);\
	stq	s4, 0x38(sp);\
	stq	s5, 0x40(sp);\
	stq	s6, 0x48(sp);	

#define load\
	ldq	gp, 0x00(sp);\
	ldq	pv, 0x08(sp);\
	ldq	ra, 0x10(sp);\
	ldq	s0, 0x18(sp);\
	ldq	s1, 0x20(sp);\
	ldq	s2, 0x28(sp);\
	ldq	s3, 0x30(sp);\
	ldq	s4, 0x38(sp);\
	ldq	s5, 0x40(sp);\
	ldq	s6, 0x48(sp);\
	addq	sp, 0x50, sp;

	
/*
**      a0      - receiver id
**
**      a1      - snd_msg_vect
**
**      a2      - rec_msg_vect
**
**      a3      - msg_timeout = (receive = 63..32 | send = 31..00)
**
**      t0..t7  - msg0 .. msg7
*/

/*	
l4_msgdope_t
l4_alpha_ipc_call(l4_threadid_t dest, 
		  const void *snd_msg,
		  const l4_ipcregs_t *snd_regs,
		  void *rcv_msg, 
		  l4_ipcregs_t *rcv_regs,
		  l4_timeout_t timeout);
*/
	
l4_alpha_ipc_call:
	save
	
	beq	a2, 1f
	ldq	t0, 0x00(a2)
	ldq	t1, 0x08(a2)
	ldq	t2, 0x10(a2)
	ldq	t3, 0x18(a2)
	ldq	t4, 0x20(a2)
	ldq	t5, 0x28(a2)
	ldq	t6, 0x30(a2)
	ldq	t8, 0x38(a2)
1:

	subq	sp, 8, sp
	stq	a4, 0(sp)
	
	mov	a3, a2
	mov	a5, a3
	bic	a2, 1, a2

	pal	L4_IPC_ENTRY

	ldq	a4, 0(sp)
	addq	sp, 8, sp

	beq	a4, 1f
	stq	t0, 0x00(a4)
	stq	t1, 0x08(a4)
	stq	t2, 0x10(a4)
	stq	t3, 0x18(a4)
	stq	t4, 0x20(a4)
	stq	t5, 0x28(a4)
	stq	t6, 0x30(a4)
	stq	t8, 0x38(a4)
1:		
	load
	
	ret	(ra)


	
/*
l4_msgdope_t
l4_alpha_ipc_reply_and_wait(l4_threadid_t dest, 
			    const void *snd_msg, 
			    const l4_ipcregs_t *snd_regs,
			    l4_threadid_t *src,
			    void *rcv_msg, 
			    l4_ipcregs_t *rcv_regs,
			    l4_timeout_t timeout);
*/
l4_alpha_ipc_reply_and_wait:
	save
	lda	fp, 0x50(sp)
	
	beq	a2, 1f
	ldq	t0, 0x00(a2)
	ldq	t1, 0x08(a2)
	ldq	t2, 0x10(a2)
	ldq	t3, 0x18(a2)
	ldq	t4, 0x20(a2)
	ldq	t5, 0x28(a2)
	ldq	t6, 0x30(a2)
	ldq	t8, 0x38(a2)
1:

	subq	sp, 16, sp
	stq	a5, 0(sp)	// rcv_regs
	stq	a3, 8(sp)	// *src
	
	mov	a4, a2		// rcv_msg
	ldq	a3, 0(fp)	// timeout
	or	a2, 1, a2

	pal	L4_IPC_ENTRY

	
	ldq	a5, 0(sp)	
	ldq	a3, 8(sp)
	addq	sp, 16, sp

	stq	pv, 0(a3)

	beq	a5, 1f
	stq	t0, 0x00(a5)
	stq	t1, 0x08(a5)
	stq	t2, 0x10(a5)
	stq	t3, 0x18(a5)
	stq	t4, 0x20(a5)
	stq	t5, 0x28(a5)
	stq	t6, 0x30(a5)
	stq	t8, 0x38(a5)
1:		
	load
	
	ret	(ra)

/*		
l4_msgdope_t
l4_i386_ipc_reply_deceiting_and_wait(const l4_ipc_deceit_ids_t ids,
				     const void *snd_msg, 
				     dword_t snd_dword0, dword_t snd_dword1,
				     l4_threadid_t *src,
				     void *rcv_msg, 
				     dword_t *rcv_dword0, dword_t *rcv_dword1, 
				     l4_timeout_t timeout, 
				     l4_msgdope_t *result);
*/

/*	
l4_msgdope_t
l4_alpha_ipc_send(l4_threadid_t dest, 
		  const void *snd_msg,
		  const l4_ipcregs_t *snd_regs,
		  l4_timeout_t timeout);
*/
l4_alpha_ipc_send:
	save
	
	beq	a2, 1f
	ldq	t0, 0x00(a2)
	ldq	t1, 0x08(a2)
	ldq	t2, 0x10(a2)
	ldq	t3, 0x18(a2)
	ldq	t4, 0x20(a2)
	ldq	t5, 0x28(a2)
	ldq	t6, 0x30(a2)
	ldq	t8, 0x38(a2)
1:

	subq	zero,1,a2

	pal	L4_IPC_ENTRY

	load
	ret	(ra)
	
/*
l4_msgdope_t
l4_alpha_ipc_wait(l4_threadid_t *src,
		  void *rcv_msg,
		  l4_ipcregs_t *rcv_regs,
		  l4_timeout_t timeout);
*/

l4_alpha_ipc_wait:
	save
	
	subq	sp, 16, sp
	stq	a2, 0(sp)
	stq	a0, 8(sp)

	or	a1, 1, a2	/* open rcv_msg */
	subq	zero, 1, a1	/* snd_msg */


	pal	L4_IPC_ENTRY

	ldq	a2, 0(sp)
	ldq	a0, 8(sp)

	addq	sp, 16, sp

	stq	pv, 0(a0)	
	beq	a2, 1f
	stq	t0, 0x00(a2)
	stq	t1, 0x08(a2)
	stq	t2, 0x10(a2)
	stq	t3, 0x18(a2)
	stq	t4, 0x20(a2)
	stq	t5, 0x28(a2)
	stq	t6, 0x30(a2)
	stq	t8, 0x38(a2)
	
1:
	load
	ret	(ra)

	

/*
l4_msgdope_t
l4_alpha_ipc_receive(l4_threadid_t src,
		     void *rcv_msg,
		     l4_ipcregs_t *rcv_regs,
		     l4_timeout_t timeout);
*/

l4_alpha_ipc_receive:
	save
	
	subq	sp, 8, sp
	stq	a2, 0(sp)

	or	a1, 0, a2	/* closed rcv_msg */
	subq	zero, 1, a1	/* snd_msg */


	pal	L4_IPC_ENTRY
	
	ldq	a2, 0(sp)

	addq	sp, 8, sp

	beq	a2, 1f
	stq	t0, 0x00(a2)
	stq	t1, 0x08(a2)
	stq	t2, 0x10(a2)
	stq	t3, 0x18(a2)
	stq	t4, 0x20(a2)
	stq	t5, 0x28(a2)
	stq	t6, 0x30(a2)
	stq	t8, 0x38(a2)
1:
	load
	ret	(ra)


	
	

	