/*
**	Copyright (c) 1995, Sebastian Schoenberg 
**	E-mail : ss10@irz.inf.tu-dresden.de
**
**	University of Technology Dresden, Department of Computer Science
**	Institute for Operating Systems, Databases and Computer Networks
**
**	$Id: thread.S,v 4.2 1998/04/17 11:02:20 schoenbg Exp $
**
*/

#if !defined(lint)
	.data
	.asciz	"$Id: thread.S,v 4.2 1998/04/17 11:02:20 schoenbg Exp $"
#endif


#include <pal/predef.h>
	
#include <pal/dc21164.h>
#include <pal/macros.h>
#include <pal/regdef.h>
#include <pal/platform.h>
	
#include <pal/l4pal.h>

#include <pal/debug.h>

#include <pal/gasp.h>
	
	.globl	lthread_ex_regs
	.globl	task_new
	.globl	thread_schedule

	.globl	create_sigma0
	
	.globl	k_switch_thread
	.globl	k_simple_switch

	.text

#define BREAK_SPACE

#ifdef DBG_ERROR	
	.globl	thread_error

thread_error:	
	sys_halt	0xeeee	
#endif

	
							
/*------------------------------------------------------------------------------
** FUNCTION: k_thread_switch
**	switches to another thread.
**
** MODE: K called from interrupt !
**
** INPUT PARAMETERS:  
**
** OUTPUT PARAMTERES:
** 
** SIDE EFFECTS:	
**	all threads are stored button-up sorted by their priority.
**	hence, HERE we don't need to start from the highest priority since
**	no thread with MCP > current MCP is activatable.
**
** C-CODE:	
**	maxprio = current->priority;
**	for (n = current->next; ; n = n->prev) {
**		if (n->priority < maxprio) {
**			maxprio = MCP_LOW; 
**			n = dispatcher->next;
**			continue;
**		}
**	}
**
*/

	
	
	
	ALIGN_BRANCH
k_switch_thread:
	push	p_all
	tcb	t0

	enqueue_busy (t0, t1, t2, t3)

	pal_addr t2, thread_priorities
        ldl_a   t1, TCB_SCHED_SCP(t0)
        s4addq  t1, t2, t2
        stl_p   t0, 0(t2)
	
		
	ldiq	t7, MAX_PRIOS+1
	pal_addr t8, (thread_priorities+((MAX_PRIOS+1)*4))

	REPEAT
		EXITZ	t7
		subq	t8, 4, t8
		subq	t7, 1, t7
		ldl_p	t0, 0(t8)
		CONTZ	t0

		ldq	t0, TCB_BUSY_NEXT (t0)
		REPEAT
			ldq	t1, TCB_THREAD_STATE (t0)	
			bic	t1, TFS_RUNNING, t1	
			IFZ	t1					// scan present_queue until found RUNNING thread
				tcb	t2
				xor	t0, t2, t2			// DON'T ACTIVATE MYSELF !
				beq	t2, switch_preempted	

				tcb_dispatcher t1
				cmpeq	t1, t0, t1
				IFZ	t1	
					switch_context t0, t10
				ENDIF
				switch_thread_kernel t0, switch_preempted	
			XENDIF
			dequeue_busy (t0, t1, t2, t3)
			mov	t0, t1
			ldl_p	t0, 0(t8)
			CONT	t0
		ENDR
		CONT	t7
	ENDR	
	kmsg	"Scheduling Error - No Thread Ready\n\r"
	debug
	halt	
	
	ALIGN_BRANCH
switch_preempted:
	pop	p_all
	pop	p_gp					// pop gp since it has been stored BEFORE kernel call
switch_activated:
#ifdef BREAK_SPACE
	push	p_t6|p_t7
	
        InPortByte(COM1_LSR,t6,t7)

	blbc	t6, 9f
	mb
	mb
        InPortByte(COM1_THR,t6,t7)
	
	cmpeq	t6, 0x20, t6
	beq	t6, 9f
	debug
9:
	pop	p_t6|p_t7
#endif	
	return_frame


/*------------------------------------------------------------------------------
** FUNCTION: k_thread_switch
**	switches to another thread.
**
** MODE: P
**
** INPUT PARAMETERS:  
**
** OUTPUT PARAMTERES:
** 
** SIDE EFFECTS
**
*/
	ALIGN_BRANCH
k_simple_switch:
	tcb	t0

	enqueue_busy (t0, t1, t2, t3)		// If not in queue, insert

	pal_addr t2, thread_priorities
        ldl_a   t1, TCB_SCHED_SCP(t0)
        s4addq  t1, t2, t2
        stl_p   t0, 0(t2)
	
	tcb_dispatcher	t1

	IF	a0
		tcb_ptr	a0, t2
		
		ldq_a	t1, TCB_MYSELF (t2)
		cmpeq	t1, a0, t1			// Is a0 a valid thread ID ?
		IF	t1				// no, so take the first one of busy queue.

			xor	t2, t0, t0			// Target == current ?
			beq	t0, switch_activated
	
			ldq	t1, TCB_THREAD_STATE (t2)
			bic	t1, TFS_RUNNING, t1		// Is the current one runnable ?
			IF	t1			// no, so take first one from the busy queue
				switch_context t2, t10
				switch_thread_kernel t2, switch_activated
			XENDIF
		ENDIF		
	ENDIF	

	
	ldiq	t7, MAX_PRIOS+1
	pal_addr t8, (thread_priorities+((MAX_PRIOS+1)*4))

	REPEAT
		EXITZ	t7

		subq	t8, 4, t8
		subq	t7, 1, t7
		ldl_p	t0, 0(t8)
		CONTZ	t0

		ldq	t0, TCB_BUSY_NEXT (t0)
		REPEAT
			ldq	t1, TCB_THREAD_STATE (t0)	
			bic	t1, TFS_RUNNING, t1	
			IFZ	t1					// scan present_queue until found RUNNING thread
				tcb	t2
				xor	t0, t2, t2			// DON'T ACTIVATE MYSELF !
				beq	t2, switch_activated	

				tcb_dispatcher t1
				cmpeq	t1, t0, t1
				IFZ	t1	
					switch_context t0, t10
				ENDIF
				switch_thread_kernel t0, switch_activated	
			XENDIF
			dequeue_busy (t0, t1, t2, t3)
			mov	t0, t1
			ldl_p	t0, 0(t8)
			cmpeq	t0, t1, t2
			IF	t2
				IF	t0
					kmsg	"t0 == t1"
					debug
				ENDIF
			ENDIF
			CONT	t0
		ENDR
		CONT	t7
	ENDR	
	kmsg	"Scheduling Error - No Thread Ready in yield\n\r"
	halt	


/*------------------------------------------------------------------------------
** FUNCTION: scheduler
**
** MODE:
**
** INPUT PARAMETERS:  
**	a0 :	thread_id
**	a1 :	preempter
**	a2 :	parameter
**	
** OUTPUT PARAMTERES:
**	t0 :	partner_id
**	t1 :	old_preempter
**	t2 :	old_parameter
**	t3 :	time_slice
** 
** SIDE EFFECTS
**
*/
	ALIGN_BRANCH
thread_schedule:
	open_frame
	tcb_ptr	a0, v0
	ldq_a	p0, TCB_MYSELF (v0)
	cmpeq	p0, a0, p0
	IF	p0

		// ptest
		// prio test valid:
		ldl_a	p0, TCB_SCHED_MCP (v0)
		tcb	t0
		ldl_a	p1, TCB_SCHED_MCP (t0)			// Intel lte is a BUG ?
		cmpule	p0, p1, p1				// is target MCP <= current MCP
		IFN	p1
			ldiq	t1, -1				// sched param is -1 on error
			close_frame
		ENDIF
		IFZ	p0					// MCP == 0, then not controlable
			ldiq	t1, -1				// sched param is -1 on error
			close_frame
		ENDIF

		ldq_a	t0, TCB_COMM_PARTNER (v0)
		ldq_a	t1, TCB_SCHED_PARAM (v0)
		ldq_a	t2, TCB_PREEMPTER (v0)

		addq	a1, 1, p0
		IF	p0
			stq_a	a1, TCB_PREEMPTER (v0)
		ENDIF
		addq	a2, 1, p0
		IF	p0
			tcb	t4
			ldl_a	t4, TCB_SCHED_MCP (t4)			// current MCP
			and	a2, 0xff, t3
			cmplt	t4, t3, t4				// cnt MCP < new SCP
			IFN	t4
				dequeue_busy (v0, p0, p1, p2)
				stl_a	t3, TCB_SCHED_SCP (v0)		// includes priority
				enqueue_busy (v0, p0, p1, p2)
			ENDIF
		ENDIF
		ldq_a	v0, TCB_SCHED_ACCOUNT (v0)
	
	ENDIF
	close_frame


	
	
	
		
/*------------------------------------------------------------------------------
** FUNCTION: task_new
**	create or allocate new task / TCB structure
**
** MODE: K
**
** INPUT PARAMETERS:  
**	a0  - task id
**	a1  - new chief / MCP
**	a2  - pager id
**	a3  - intital IP
**
** OUTPUT PARAMTERES:
** 
** SIDE EFFECTS
**
*/
/*
**
** if permitted
**	if ptab-entry >= 0 then
**		delete all subtasks
**		delete this task
**	endif
**
**	if pager is NIL then
**		ptab-entry = tcb of task // less than zero
**	else
**		if ptab-entry is task of current tcb then
**			create active task
**		endif
** 	endif
** endif
**
*/
	ALIGN_BRANCH
task_new:
	open_frame
	kernel	task_new_kern
	
	ALIGN_BRANCH
task_new_kern:
	push	p_ra

	tcb	t0
	tcb_dispatcher t1
	xor	t0, t1, t1					// dispatcher can do everything.
	IF	t1
		ldq_a	t1, TCB_MYSELF (t0)			// get myself id
#ifdef CHIEFTASK
		srl	a0, TID_S_CHIEF, t2
		srl	t1, TID_S_TASK, t3
		xor	t2, t3, t2
		ldiq	AT, TID_M_CHIEF				// Chief Mask	
		and	t2, AT, AT				// AT is 0, if chief of a0 == myself
		bne	AT, tn_failed_permission
#endif
		xor	a0, t1, t2
		srl	t2, TID_S_SITE, t2
		ldiq	AT, TID_M_SITE
		and	t2, AT, AT
		bne	AT, tn_failed_permission		// Failed if different site given
	ENDIF

	// Well, permissions are ok, now we have the right to modify
	// the task
	task_nr	a0, t0						// Get Task nr of given id

	pal_addr t1, pal_ptroots				// root pointer of ptabs
	s8addq	t0, t1, t0					// page base pointer to.
	ldl_p	t1, 0(t0)					// t0 now the page base	

	IFGE	t1						// there is a page table entry
		bsr	ra, task_delete				// first delete that task including subtasks
	ENDIF

	ldl_p	t1, 0(t0)		
	tcb	t2

	xor	t2, t1, t2
	srl	t2, TID_S_TASK, t2				// task doesn't have the right to create task

	IF	t2
		kmsg	"No permission to create task\n\r"
		pop	p_ra
		clr	v0
		return_frame
	XENDIF

	// Create now the address space from scratch
	IF	a2						// pager given, create active task
		bsr	ra, mem_grab_frame
	
		mov	v0, t0
		pal_addr t1, pal_ptbase

		ldiq	t2, 8
		REPEAT						// Copy the initial kernel page table
			ldq_p	AT, 0(t1)
			addq	t1, 8, t1
			subq	t2, 1, t2
			stq_p	AT, 0(t0)
			addq	t0, 8, t0
			CONT	t2	
		ENDR
		addq	v0, 64-2, v0
		task_nr	a0, t0					// Get Task nr

		pal_addr t1, pal_ptroots			// root pointer of ptabs
		s8addq	t0, t1, t0				// page base pointer t0.
		stl_p	v0, 0(t0)				// v0 now to the page base


		ldiq	AT, TID_M_THREAD
		bic	a0, AT, a0				// a0 is now task id
		tcb_ptr	a0, v0

		clr	a4					// No sp given

		tcb	t1
		ldl	t2, TCB_SCHED_MCP (t1)
		and	a1, 0xff, a1				// MCP
		cmplt	t2, a1, t3				// cnt MCP < new MCP
		cmovne	t3, t2, a1				// yes a1 is cnt MCP
	
		inswl	a1, 4, a1				// MCP
		ldl	t2, TCB_SCHED_SCP (t1)
		or	a1, t2, a5				// Same SCP as current

	
		bsr	ra, thread_fill_tcb	
	ELSE							// no pager -> grant this task
		tcb_ptr	a1, t1					// now this task is inactive
		stl_p	t1, 0(t0)				
		clr	v0					// now the new task can create this task active
	ENDIF
	pop	p_ra
	return_frame


/*
** delete the task including all subtasks
** t0 is pointer to pt_root entry
*/	
task_delete:
	tcb	t1					// now the task right belongs to the the deleter
	stl_p	t1, 0(t0)					
	kmsg	"Task deletion not yet implemented\n\r"
	ret	(ra)

	
tn_failed_permission:
	kmsg	"Access denied\n\r"
	pop	p_ra
	clr	v0
	return_frame
	


/*------------------------------------------------------------------------------
** FUNCTION: lthread_ex_regs
**
** MODE: U
**
** INPUT PARAMETERS:  
**	a0 :	thread_number
**	a1 :	preempter
**	a2 :	pager
**	a3 :	pc
**	a4 :	sp
**
** OUTPUT PARAMTERES:
** 
** SIDE EFFECTS
**
*/
	ALIGN_BRANCH
lthread_ex_regs:
	open_frame
	kernel	lthread_ex_regs_kern

	ALIGN_BRANCH
lthread_ex_regs_kern:		
	push	p_ra
	sra	sp, TID_S_TASK, t0
	and	a0, TID_M_THREAD, a0
	sll	t0, TID_S_TASK, t0
	sll	a0, TID_S_THREAD, a0
	or	a0, t0, v0

	ldl	t2, TCB_COARSE_STATE (v0)			// Set Used bit in coarse state	
	IFLBC	t2
		addq	a3, 1, t0				// if a3 or a4 == -1 than exit
		IF	t0
			addq	a4, 1, t0
			IF	t0
				tcb	t0
				ldq	a5, TCB_SCHED_PARAM (t0)// Gets same prio current thread
				bsr	ra, thread_fill_tcb	// if this TCB is not used, fill it.
				pop	p_ra
				return_frame
			XENDIF
		ENDIF		
		clr	a1
		clr	a2
		clr	a3

		pop	p_ra
		return_frame
		
	ENDIF
	addq	a1, 1, t0
	ldq	t1, TCB_PREEMPTER (v0)
	IF	t0
		stq	a1, TCB_PREEMPTER (v0)			// Save Preempter if not invalid
	ENDIF
	mov	t1, a1

	addq	a2, 1, t0
	ldq	t1, TCB_PAGER (v0)
	IF	t0
		stq	a2, TCB_PAGER (v0)			// Save pager, if not invalid
	ENDIF
	mov	t1, a2

	lda	t1, TCB_STACKTOP (v0)				// User SP and PC are always ontop
								// of the user stack
	addq	a3, 1, t0
	ldq	t2, -24(t1)
	IF	t0
		stq	a3, -24(t1)				// Load user mode pc
	ENDIF
	mov	t2, a3
	
	addq	a4, 1, t0
	ldq	t2, -8(t1)
	IF	t0
		stq	a4, -8(t1)					// Load old user mode ip
		mov	t2, a4
		/* now, there may be a in-kernel IPC operation that is going to be
		** terminated - otherwise, simply return
		*/
		ldq	t0, TCB_THREAD_STATE (v0)
		bic	t0, (TFS_LOCKED_WAITING|TFS_LOCKED_RUNNING), t1
		IFZ	t1						// state was locked somewhat 
			pal_addr t2, ipc_ret_abort+1
			lda	t1, (TCB_STACKTOP-4*8) (v0)
			stq	t0, 0(t1)					// "push" the old thread state
			br	ex_kernel
		ENDIF
		bic	t0, (TFS_POLLING), t1
		IFZ	t1
			pal_addr t2, ipc_ret_cancel_polling+1		// state was polling
			lda	t1, (TCB_STACKTOP-3*8) (v0)
			br	ex_kernel
		ENDIF
		bic	t0, (TFS_ABORTED), t1
		IF	t1
			pal_addr t2, ipc_ret_cancel_wait+1		// state was receiving (open/close)
			lda	t1, (TCB_STACKTOP-3*8) (v0)
			br	ex_kernel
		ENDIF
		pop	p_ra
		return_frame
	
	XENDIF
	mov	t2, a4
	pop	p_ra
	return_frame
	
ex_kernel:
	stq	t1, TCB_KSP (v0)
	
	stq	t2, TCB_RESTART (v0)

	ldiq	t2, TFS_RUNNING
	stq	t2, TCB_THREAD_STATE (v0)

	enqueue_busy (v0, t1, t2, t3)
	pop	p_ra
	return_frame
	
/*	
** Create new thread, strictly speaking, fill the TCB
**  
**	v0 - TCB address
**	a1 - preemptor
**	a2 - pager
**	a3 - ip
**	a4 - sp
**	a5 - TCB_SCHED_PARAM
**
** RETURN:	
**	v0 - THREAD_ID
**			
*/		 
thread_fill_tcb:
	ldiq	t2, TCS_USED
	stl	t2, TCB_COARSE_STATE (v0)		// Set Used bit in coarse state

	ldiq	t2, VER_TCB_ALPHA	
	stl	t2, TCB_IDENT (v0)			// Mark TCB(v0) as valid

	stq	a1, TCB_PREEMPTER (v0)
	stq	a2, TCB_PAGER (v0)
	
	lda	t2, TCB_SEND_ROOT (v0)			// Init send queue
	stq	t2, TCB_SEND_ROOT (v0)
	stq	t2, (TCB_SEND_ROOT+8) (v0)

	stq	zero, TCB_WAKEUP_QUEUE (v0)
	

	ldiq	t2, TFS_RUNNING				// Set new thread running
	stq	t2, TCB_THREAD_STATE (v0)

	stq	a5, TCB_SCHED_PARAM(v0)
		
	tcb	t2
	
	ldq	t4, TCB_MYSELF (v0)
	sll	v0, 35, t3				// Fabrique my TID
	srl	t3, 35, t3				// Only task_no and thread_no now
	addq	t4, 1, t4
	or	t3, t4, t4				// Merge with the ver bits
	stq	t4, TCB_MYSELF (v0)			// t4 = thread id
			
	disable_int
	tcb_dispatcher	t2
	insert_ll (v0, t2, t3, TCB_PRESENT_QUEUE)	// Insert v0 into present list after t2, user t3 as temp

	stq	v0, TCB_BUSY_QUEUE (v0)		// Initialize busy queue
	stq	v0, (TCB_BUSY_QUEUE+8)(v0)
	

	bis	zero, LLS_PRESENT_QUEUE, t2
	stl	t2, TCB_LIST_STATE (v0)			// Set list state

	lda	t2, TCB_STACKTOP (v0)			// size of TCB - stack top

	lda	t1, _thread_startup			// Address of schedule in point
	subq	t2, 24, t2
	stq	a4, 16(t2)				// +16 stack	 

	subq	zero, 1, t0				// Set to user mode
	stq	t0, 8(t2)				// +8 mode

	bic	a3, 3, a3				// No PAL please !
	
	stq	a3, 0(t2)				// +0 user address
	stq	t2, TCB_KSP (v0)			// set new KSP
	stq	t1, TCB_RESTART (v0)

	enqueue_busy (v0, t1, t2, t3)
	mov	t4, v0
	enable_int
	ret	zero, (ra)



/*------------------------------------------------------------------------------
** FUNCTION: _thread_startup
**	loads the user address from kernel stack and starts that
**	code
**
** MODE: K
**
** SIDE EFFECTS:
**	currently, sets pv to the address of the procedure
**	that allows a simpler call of gcc generated func-
**	tions.
*/
	ALIGN_BRANCH
_thread_startup:
	tcb	a1
	stq_a	zero, TCB_RESTART (a1)
	
	nop
	nop
	ldq	pv, 0(sp)				// load procedure value register
	nop
	return_frame					// this is a gimick...

/*--------------------------------------------------------------
** vu2
** FUNCTION: l4_id_nearest
**	returns the nearest id of a chief for a given id of a task, or 
**	the id of the thread itself
**
** INPUT PARAMETERS:
**
**	a0 : NULL / dest-id
**
** OUTPUT PARAMETERS:
**
**	v0 : id
**
** INTERNAL:
**
**	t0 : id_myself
**	t1 : id_dest
**	t2 : tcb_myself
**	t3 : tcb_dest
*/

/*
** C-code:
** 	while (myself->depth < dest->depth)
**		dest=dest->chief;
**	if (myself->chief_id == dest->chief_id) return dest_id;
**	else return myself->chief_id;
*/
	.globl	l4_id_nearest		
l4_id_nearest:
	tcb 	t0
	bis 	zero, a0, t1			
	tcb_ptr	t0, t2
	tcb_ptr	t1, t3
	REPEAT
		ldq	t4, TCB_MYSELF(t2)
		srl	t4, TID_S_DEPTH, t4			// preparing the depth
//		and	t4, TID_M_DEPTH, t4			// into t4
		ldq	t5, TCB_MYSELF(t3)
		srl	t5, TID_S_DEPTH, t5			// and the other 
//		and	t5, TID_M_DEPTH, t5			// is already 0
		cmpule	t4, t5, t6				// depth_dest>depth_myself
		IFZ 	t6
			// a loop, where the destination depth becomes the same like myself!
			ldq 	t5, TCB_MYSELF(t3)	
			srl	t5, TID_S_TASK, t1
			ldiq	t3, TID_M_TASK
			and	t1, t3, t1
			sll	t1, TID_S_THREAD, t1		// the new thread-id
			tcb_ptr	t1, t3
			CONT
		ENDIF
	ENDR
	ldq_a	t4, TCB_MYSELF(t2)
	ldq_a	t5, TCB_MYSELF(t3)
	xor	t4, t5, t6
	srl	t6, TID_S_CHIEF, t6
		
	cmovne	t6, t5, v0
	IF	t6	
		sll	t6, TID_S_THREAD, t6			// the chief in t6 is the nearest
		tcb_ptr t6, v0
		ldq	v0, TCB_MYSELF(v0)
	ENDIF
	return_frame
	

/*------------------------------------------------------------------------------
** FUNCTION: create_sigma0
**	create the address space etc. for sigma 0
**
** MODE: K
**
** INPUT PARAMETERS:  
**
** OUTPUT PARAMTERES:
** 
** SIDE EFFECTS
**
*/
	
create_sigma0:
	push	p_ra
	ldiq	a0, (TASK_SIGMA0 << TID_S_TASK)
	ldiq	a1, MCP_HIGH			
	subq	zero, 1, a2
	ldiq	a3, P_SIGMA0BASE
	pal	L4_TASK_NEW_ENTRY

	mov	v0, a0				// thread id
	ldiq	a1, -1
	ldiq	a2, SCP_NORMAL			// for fun... 
	pal	L4_SCHEDULE_ENTRY

	addq	t1, 1, t0
	IFZ	t0
		kmsg	"Cannot set Sigma0 prio ?"
		debug
	ENDIF
	
	pop	p_ra
	ret	zero, (ra)


	.data
		
MSG_task_in_use:
	.asciz	"TCB is in use\n\r"

	
	.align	3
.END








