/*
**
**	L4 Alpha Kernel
**	Copyright (c) 1996, Sebastian Schoenberg
**	University of Technology Dresden, University of Cambridge
**
*/



#include <l4/types.h>
#include <l4/syscalls.h>
#include <l4/ipc.h>

#include <l4/kdebug.h>

qword_t timeout_stack [1024];

#define PAGEMASK 0x1fff


inline void l4_grant_task (l4_threadid_t task, l4_taskid_t destination)
{
  l4_task_new (task, destination.thread_id, 0, L4_NIL_ID);
}

#define SIGMA0_TASK	1	
#define USER0_TASK	2

int main (void)
{
  l4_threadid_t myself, from, oid;
  l4_taskid_t tid, des;
  qword_t i, old = -1;
  l4_ipcregs_t regs;  
  l4_msgdope_t retn;

  print_string ("Sigma 0 startup\n\r");

  des = tid = myself = l4_myself ();
  tid.id.task = USER0_TASK;

  /* Set new - rights for new task to create tasks */
  
  for (i = USER0_TASK+1; i < 1024; i++) {
    des.id.task = i;
    l4_grant_task (des, tid);
  }
 
  /* Maximum CP < current CP */ 
  l4_task_new (tid, 0xf0, 0x222000, myself);

  print_string ("Sigma 0 is requesting page faults\n\r");

  l4_alpha_ipc_wait (&from,
		     NULL,
		     &regs,
		     L4_IPC_NEVER);

  while (1) {

    if ((regs.val[0] == old) && (oid.id.task == from.id.task)) {
      print_string ("Got addr 2 times\n\rAddr "); print_hex (old, 64); 
      print_string (";  IP = "); print_hex (regs.val[1], 64);
      print_string ("\n\rFaulting Thread is ");
      print_hex (oid.thread_id, 64); print_string ("\n\r");
    }
    old = regs.val[0]; oid = from;
    regs.val[0] = 
      regs.val[1] = (regs.val[0] & ~PAGEMASK) | (13 << 2) | 1;


#ifdef SINGLE    
    retn = l4_alpha_ipc_send (from, 
		       (void*)((qword_t)(NULL) | 0x02),
		       &regs,
		       L4_IPC_NEVER);
    if (retn.msgdope) {
      print_string ("Cannot send, why ?\n\r");
      print_hex (retn.msgdope, 64);
    }

    retn = l4_alpha_ipc_wait (&from,
		       NULL,
		       &regs,
		       L4_IPC_NEVER);
    if (retn.msgdope) {
      print_string ("Cannot wait, why ?\n\r");
      print_hex (retn.msgdope, 64);
    }
#else

    retn = l4_alpha_ipc_reply_and_wait (from,
					(void*)((qword_t)(NULL) | 0x02),
					&regs,

					&from,
					NULL,
					&regs,

					L4_IPC_NEVER);
    if (retn.msgdope) {
      print_string ("Cannot send, why ?\n\r");
      print_hex (retn.msgdope, 64);
    }
#endif
  }
}






